"""
Shell command generation for
- running queries in databases via their command line clients
- copying data from, into and between databases
"""

import shlex
from functools import singledispatch
from warnings import warn

import sys
from mara_db import dbs, config
from multimethod import multidispatch

from mara_db import formats
from mara_db.formats import _check_format_with_args_used, _get_format_from_args


@singledispatch
def query_command(db: object, timezone: str = None, echo_queries: bool = None) -> str:
    """
    Creates a shell command that receives a sql query from stdin and executes it

    Args:
        db: The database in which to run the query (either an alias or a `dbs.DB` object
        timezone: Sets the timezone of the client, if applicable
        echo_queries: Whether the client should print executed queries, if applicable

    Returns:
        A shell command string

    Examples:
        >>> print(query_command('mara', 'America/New_York'))
        PGTZ=America/New_York PGOPTIONS=--client-min-messages=warning psql --username=root --host=localhost \
            --echo-all --no-psqlrc --set ON_ERROR_STOP=on mara


        >>> print(query_command(dbs.MysqlDB(host='localhost', database='test')))
        mysql --default-character-set=utf8mb4 --host=localhost test
    """
    raise NotImplementedError(f'Please implement query_command for type "{db.__class__.__name__}"')


@query_command.register(str)
def __(alias: str, timezone: str = None, echo_queries: bool = None):
    return query_command(dbs.db(alias), timezone=timezone, echo_queries=echo_queries)


@query_command.register(dbs.PostgreSQLDB)
def __(db: dbs.PostgreSQLDB, timezone: str = None, echo_queries: bool = None):
    if echo_queries is None:
        echo_queries = config.default_echo_queries()
    if timezone is not None:
        warn("Parameter timezone is deprecated and will be removed in mara-db 5.0, see https://github.com/mara/mara-db/issues/44")

    return (f'PGTZ={timezone or config.default_timezone()} '
            + (f"PGPASSWORD='{db.password}' " if db.password else '')
            + (f'PGSSLMODE={db.sslmode} ' if db.sslmode else '')
            + (f'PGSSLROOTCERT={db.sslrootcert} ' if db.sslrootcert else '')
            + (f'PGSSLCERT={db.sslcert} ' if db.sslcert else '')
            + (f'PGSSLKEY={db.sslkey} ' if db.sslkey else '')
            + 'PGOPTIONS=--client-min-messages=warning psql'
            + (f' --username={db.user}' if db.user else '')
            + (f' --host={db.host}' if db.host else '')
            + (f' --port={db.port}' if db.port else '')
            + (' --echo-all' if echo_queries else ' ')
            + ' --no-psqlrc --set ON_ERROR_STOP=on '
            + (db.database or ''))


@query_command.register(dbs.RedshiftDB)
def __(db: dbs.RedshiftDB, timezone: str = None, echo_queries: bool = None):
    if echo_queries is None:
        echo_queries = config.default_echo_queries()
    if timezone is not None:
        warn("Parameter timezone is deprecated and will be removed in mara-db 5.0, see https://github.com/mara/mara-db/issues/44")

    return (f'PGTZ={timezone or config.default_timezone()} '
            + (f"PGPASSWORD='{db.password}' " if db.password else '')
            + ' psql'
            + (f' --username={db.user}' if db.user else '')
            + (f' --host={db.host}' if db.host else '')
            + (f' --port={db.port}' if db.port else '')
            + (' --echo-all' if echo_queries else ' ')
            + ' --no-psqlrc --set ON_ERROR_STOP=on '
            + (db.database or ''))


@query_command.register(dbs.BigQueryDB)
def __(db: dbs.BigQueryDB, timezone: str = None, echo_queries: bool = None):
    from .bigquery import bigquery_credentials

    service_account_email = bigquery_credentials(db).service_account_email

    return (f'CLOUDSDK_CORE_ACCOUNT={service_account_email}'
            + ' bq query'
            + ' --max_rows=50000000'  # run without user interaction
            + ' --headless'  # run without user interaction
            + ' --quiet'  # no job progress updates
            + ' --format=csv'  # no job progress updates
            + (' --use_legacy_sql=' + ('true' if db.use_legacy_sql else 'false'))
            + (f' --project_id={db.project}' if db.project else '')
            + (f' --location={db.location}' if db.location else '')
            + (f' --dataset_id={db.dataset}' if db.dataset else '')
            + ' ')


@query_command.register(dbs.MysqlDB)
def __(db: dbs.MysqlDB, timezone: str = None, echo_queries: bool = None):
    assert all(v is None for v in [timezone, echo_queries]), "unimplemented parameter for MysqlDB"
    return ((f"MYSQL_PWD='{db.password}' " if db.password else '')
            + 'mysql '
            + (f' --user={db.user}' if db.user else '')
            + (f' --host={db.host}' if db.host else '')
            + (f' --port={db.port}' if db.port else '')
            + (' --ssl' if db.ssl else '')
            + (f' {db.database}' if db.database else ''))


@query_command.register(dbs.SqshSQLServerDB)
def __(db: dbs.SqshSQLServerDB, timezone: str = None, echo_queries: bool = None):
    assert timezone is None, "unimplemented parameter for SqshSQLServerDB"

    if echo_queries is None:
        echo_queries = config.default_echo_queries()

    # sqsh does not do anything when a statement is not terminated by a ';', add one to be sure
    command = "(cat && echo ';') \\\n  | "
    command += "(cat && echo ';\n\go') \\\n  | "

    return (command + 'sqsh -a 1 -d 0 -f 10'
            + (f' -U {db.user}' if db.user else '')
            + (f' -P {db.password}' if db.password else '')
            + (f' -S {db.host}' if db.host else '')
            + (f':{db.port}' if db.host and db.port and db.port != 1433 else '')
            + (f' -D {db.database}' if db.database else '')
            + (' -e' if echo_queries else ''))


@query_command.register(dbs.SqlcmdSQLServerDB)
def __(db: dbs.SqlcmdSQLServerDB, timezone: str = None, echo_queries: bool = None):
    assert timezone is None, "unimplemented parameter for SQLServerDB"

    if echo_queries is None:
        echo_queries = config.default_echo_queries()

    if db.host:
        # connection to DB, see: https://docs.microsoft.com/en-us/sql/ssms/scripting/sqlcmd-connect-to-the-database-engine?view=sql-server-ver15
        if db.protocol == 'tcp':
            port = db.port if db.port else 1433
            server = f'tcp:{db.host},{port}'
        elif db.protocol == 'np':
            pipe = f'MSSQL${db.instance}\\sql\\query' if db.instance else 'pipe\\sql\\query'
            server = f'np:\\\\{db.host}\\{pipe}'
        elif db.protocol == 'lpc':
            server = f'lcp:{db.host}\\{db.instance}' if db.instance else f'lcp:{db.host}'
        else:
            if db.instance:
                server = f'{db.host}\\{db.instance}'
            elif db.port:
                server = f'{db.host},{db.port}'
            else:
                server = db.host

    return ('sqlcmd -b -r'
            + (f' -U {db.user}' if db.user else '')
            + (f' -P {db.password}' if db.password else '')
            + (f' -S {server}' if server else '')
            + (' -C' if db.trust_server_certificate else '')
            + (f' -d {db.database}' if db.database else '')
            + (' -e' if echo_queries else '')
            + (' -I' if db.quoted_identifier else '')
            + (' -i /dev/stdin'))


@query_command.register(dbs.OracleDB)
def __(db: dbs.OracleDB, timezone: str = None, echo_queries: bool = None):
    assert all(v is None for v in [timezone, echo_queries]), "unimplemented parameter for OracleDB"

    # sqlplus does not do anything when a statement is not terminated by a ';', add one to be sure
    return (  # Oracle needs a semicolon at the end, with no newlines before
        # Remove all trailing whitespace and then add a semicolon if not there yet
            shlex.quote(sys.executable)
            + ''' -c "import sys; sql = sys.stdin.read().strip(); sql = sql + ';' if not sql[-1]==';' else sql; print(sql)" '''
            + ' \\\n  | sqlplus64 -s '
            + f'{db.user}/{db.password}@{db.host}:{db.port or 1521}/{db.endpoint}')


@query_command.register(dbs.SQLiteDB)
def __(db: dbs.SQLiteDB, timezone: str = None, echo_queries: bool = None):
    assert all(v is None for v in [timezone, echo_queries]), "unimplemented parameter for SQLiteDB"

    # sqlite does not complain if a file does not exist. Therefore check file existence first
    file_name = shlex.quote(str(db.file_name))
    return f'(test -f {file_name} && cat || >&2 echo {file_name} not found) \\\n' \
           + '  | sqlite3 -bail ' + shlex.quote(str(db.file_name))


@query_command.register(dbs.SnowflakeDB)
def __(db: dbs.SnowflakeDB, timezone: str = None, echo_queries: bool = None):
    assert timezone is None, "unimplemented parameter for SnowflakeDB"
    assert not echo_queries, "unimplemented parameter for SnowflakeDB"
    return ((f'SNOWSQL_PWD={shlex.quote(db.password)} ' if db.password else '')
            +(f'SNOWSQL_PRIVATE_KEY_PASSPHRASE={shlex.quote(db.private_key_passphrase)}' if db.private_key_passphrase else '')
            + 'snowsql --noup -o exit_on_error=true'
            +(f' -c {db.connection}' if db.connection else '')
            +(f' -a {db.account}' if db.account else '')
            +(f' -u {db.user}' if db.user else '')
            +(f' --private-key-path {db.private_key_file}' if db.private_key_file else '')
            +(f' -d {db.database}' if db.database else '')
            +' -f /dev/stdin')


@query_command.register(dbs.DatabricksDB)
def __(db: dbs.DatabricksDB, timezone: str = None, echo_queries: bool = None):
    assert timezone is None, "unimplemented parameter for DatabricksDB"
    assert not echo_queries, "unimplemented parameter for DatabricksDB"
    return ('dbsqlcli'
            + (f' --hostname {db.host}' if db.host else '')
            + (f' --http-path {db.http_path}' if db.http_path else '')
            + (f' --access-token {db.access_token}' if db.access_token else '')
            + ' -e /dev/stdin')


# -------------------------------


@singledispatch
def copy_to_stdout_command(db: object,
                           header: bool = None,
                           footer: bool = None,
                           delimiter_char: str = None,
                           csv_format: bool = None,
                           pipe_format: formats.Format = None) -> str:
    """
    Creates a shell command that receives a query from stdin, executes it and writes the output to stdout

    Args:
        db: The database in which to run the query (either an alias or a `dbs.DB` object
        header: Whether a csv header with the column name(s) will be included or not.
            No header, by default. (not implemented in sqsh for SQLServerDB)
        footer: Whether a footer will be included or not. False by default. (Only implemented for PostgreSQLDB)
        delimiter_char: str to delimit the fields in one row. Default: tab character
        csv_format: Double quote 'difficult' strings (Only implemented for PostgreSQLDB)
        pipe_format: The format passed to stdout

    Returns:
        The composed shell command

    Example:
        >>> print(copy_to_stdout_command(dbs.PostgreSQLDB(host='localhost', database='test')))
        PGTZ=Europe/Berlin PGOPTIONS=--client-min-messages=warning psql --host=localhost  --no-psqlrc --set ON_ERROR_STOP=on test --tuples-only --no-align --field-separator='	' \
            | grep -a -v -e '^$'
    """
    raise NotImplementedError(f'Please implement function copy_to_stdout_command for type "{db.__class__.__name__}"')


@copy_to_stdout_command.register(str)
def __(alias: str, header: bool = None, footer: bool = None, delimiter_char: str = None, csv_format: bool = None,
       pipe_format: formats.Format = None):
    return copy_to_stdout_command(dbs.db(alias), header=header, footer=footer,
                                  delimiter_char=delimiter_char, csv_format=csv_format,
                                  pipe_format=pipe_format)


@copy_to_stdout_command.register(dbs.PostgreSQLDB)
def __(db: dbs.PostgreSQLDB, header: bool = None, footer: bool = None,
       delimiter_char: str = None, csv_format: bool = None, pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    if isinstance(pipe_format, formats.CsvFormat):
        assert not (pipe_format.footer or pipe_format.header), 'unsupported when format is CsvFormat'
        delimiter_char = pipe_format.delimiter_char or ','
        return (r" sed '/\;/q' | sed 's/\;.*//' " + '\\\n'
               + f'''| (echo "COPY (" && cat && echo ") TO STDOUT WITH CSV  DELIMITER '{delimiter_char}' ") \\\n'''
               + '  | ' + query_command(db, echo_queries=False) + ' --variable=FETCH_COUNT=10000 \\\n'
               + "  | sed '/^$/d'")  # remove empty lines

    elif isinstance(pipe_format, formats.NativeFormat):
        header_argument = '--tuples-only' if not header else ''
        footer_argument = '--pset="footer=off"' if not footer else ''
        return (query_command(db, echo_queries=False) + ' --variable=FETCH_COUNT=10000'
                + " " + header_argument + " " + footer_argument
                + f" --no-align --field-separator='{None}' \\\n"
                + "  | sed '/^$/d'"  # remove empty lines
                )

    else:
        raise ValueError(f'Unsupported pipe_format for PostgreSQLDB: {pipe_format}')


@copy_to_stdout_command.register(dbs.BigQueryDB)
def __(db: dbs.BigQueryDB, header: bool = None, footer: bool = None, delimiter_char: str = None,
       csv_format: bool = None, pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.header:
            raise ValueError('Unsupported pipe_format.header for BigQueryDB')
        if pipe_format.footer:
            raise ValueError('Unsupported pipe_format.footer for BigQueryDB')

    elif isinstance(pipe_format, formats.NativeFormat):
        pass

    else:
        raise ValueError(f'Unsupported pipe_format for BigQueryDB: {pipe_format}')

    remove_header = 'sed 1d'
    return query_command(db) + f' | {remove_header}'


@copy_to_stdout_command.register(dbs.MysqlDB)
def __(db: dbs.MysqlDB, header: bool = None, footer: bool = None, delimiter_char: str = None, csv_format: bool = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    header: bool

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.footer:
            raise ValueError('Unsupported pipe_format.footer for MysqlDB')
        if pipe_format.delimiter_char:
            raise ValueError('Unsupported pipe_format.delimiter_char for MysqlDB')

        header = pipe_format.header

    elif isinstance(pipe_format, formats.NativeFormat):
        header = False

    else:
        raise ValueError(f'Unsupported pipe_format for MysqlDB: {pipe_format}')

    header_argument = '--skip-column-names' if not header else ''
    return query_command(db) + ' ' + header_argument


@copy_to_stdout_command.register(dbs.SqshSQLServerDB)
def __(db: dbs.SqshSQLServerDB, header: bool = None, footer: bool = None, delimiter_char: str = None,
       csv_format: bool = None, pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.header:
            raise ValueError('pipe_format.header is not supported for SqshSQLServerDB')
        if pipe_format.footer:
            raise ValueError('pipe_format.footer is not supported for SqshSQLServerDB')
        if pipe_format.delimiter_char and pipe_format.delimiter_char != ',':
            raise ValueError(f"For SqshSQLServerDB delimiter_char must ','")

    elif isinstance(pipe_format, formats.NativeFormat):
        pass

    else:
        raise ValueError(f'Unsupported pipe_format for SqshSQLServerDB: {pipe_format}')

    return query_command(db, echo_queries=False) + " -m csv"


@copy_to_stdout_command.register(dbs.SqlcmdSQLServerDB)
def __(db: dbs.SqlcmdSQLServerDB, header: bool = None, footer: bool = None, delimiter_char: str = None,
       csv_format: bool = None, pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    header: bool
    delimiter_char: str

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.footer:
            raise ValueError('pipe_format.footer is not supported for SqlcmdSQLServerDB')

        header = pipe_format.header
        delimiter_char = pipe_format.delimiter_char

    elif isinstance(pipe_format, formats.NativeFormat):
        header = False

    else:
        raise ValueError(f'Unsupported pipe_format for SqlcmdSQLServerDB: {pipe_format}')

    # manipulate the SQL query
    command = "(echo 'SET NOCOUNT ON\n' && cat) \\\n  | "
    command += "(echo 'GO\n' && cat) \\\n  | "

    return (command + query_command(db, echo_queries=False)
            + ' -W'
            + (f''' "-s{delimiter_char or ','}"''')
            + (f' -h-1' if not header else '')
            + (f' -w 65535') # see https://docs.microsoft.com/en-us/sql/tools/sqlcmd-utility?view=sql-server-ver15, is used to avoid line-break when output is longer then 80 characters
            # removes the dashed line between header and data rows
            + (" | sed -e '2d'" if header else ''))


@copy_to_stdout_command.register(dbs.OracleDB)
def __(db: dbs.OracleDB, header: bool = None, footer: bool = None, delimiter_char: str = None, csv_format: bool = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.header:
            raise ValueError('pipe_format.header is not supported for OracleDB')
        if pipe_format.footer:
            raise ValueError('pipe_format.footer is not supported for OracleDB')
        if pipe_format.delimiter_char:
            raise ValueError('pipe_format.delimiter_char is not supported for OracleDB')

    elif isinstance(pipe_format, formats.NativeFormat):
        pass

    else:
        raise ValueError(f'Unsupported pipe_format for OracleDB: {pipe_format}')

    return "(echo 'set markup csv on\nset feedback off\nset heading off' && cat)" \
           + " \\\n  | " + query_command(db)


@copy_to_stdout_command.register(dbs.SQLiteDB)
def __(db: dbs.SQLiteDB, header: bool = None, footer: bool = None, delimiter_char: str = None, csv_format: bool = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    header_argument: str
    delimiter_char: str

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.footer:
            raise ValueError('pipe_format.footer is not supported for SQLiteDB')
        
        header_argument = '-noheader' if not pipe_format.header else ''
        delimiter_char = pipe_format.delimiter_char or '\t'

    elif isinstance(pipe_format, formats.NativeFormat):
        header_argument = ''
        delimiter_char = '\t'

    else:
        raise ValueError(f'Unsupported pipe_format for SQLiteDB: {pipe_format}')

    return query_command(db) + " " + header_argument + f" -separator '{delimiter_char}' -quote"


@copy_to_stdout_command.register(dbs.SnowflakeDB)
def __(db: dbs.SnowflakeDB, header: bool = None, footer: bool = None, delimiter_char: str = None, csv_format: bool = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    output_format: str # https://docs.snowflake.com/en/user-guide/snowsql-config.html#output-format
    header: bool

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.footer:
            raise ValueError('pipe_format.footer is not supported for SnowflakeDB')

        if pipe_format.delimiter_char == ',' or pipe_format.delimiter_char is None:
            output_format = 'csv'
        elif pipe_format.delimiter_char == '\t':
            output_format = 'tsv'
        else:
            raise ValueError(f"Not supported delimiter_char for SnowflakeDB: '{pipe_format.delimiter_char}'")

        header = pipe_format.header

    elif isinstance(pipe_format, formats.NativeFormat):
        output_format = 'csv'
        header = False

    else:
        raise ValueError(f'Unsupported pipe_format for SnowflakeDB: {pipe_format}')       

    # possible other output_format, not implemented:
    #   json = json array

    # see also: https://docs.snowflake.com/en/user-guide/snowsql-use.html#exporting-data
    return (query_command(db, echo_queries=False) + f' -o output_format={output_format} -o friendly=false -o timing=false'
            +(f' -o header=true' if header else ' -o header=false'))


@copy_to_stdout_command.register(dbs.DatabricksDB)
def __(db: dbs.DatabricksDB, header: bool = None, footer: bool = None, delimiter_char: str = None, csv_format: bool = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=header, footer=footer, delimiter_char=delimiter_char, csv_format=csv_format)

    table_format: str
    remove_header: bool

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.footer:
            raise ValueError('pipe_format.footer is not supported for DatabricksDB')

        if pipe_format.delimiter_char == ',' or pipe_format.delimiter_char is None:
            table_format = 'csv'
        elif pipe_format.delimiter_char == '\t':
            table_format = 'tsv'
        else:
            raise ValueError(f"Not supported delimiter_char for DatabricksDB: '{pipe_format.delimiter_char}'")

        if pipe_format.header:
            remove_header = True

    elif isinstance(pipe_format, formats.NativeFormat):
        table_format = 'csv'

    else:
        raise ValueError(f'Unsupported pipe_format for DatabricksDB: {pipe_format}')       

    return (query_command(db, echo_queries=False)
            + f' --table-format {table_format}'
            + ('\n  | sed 1d' if remove_header else ''))


# -------------------------------


@singledispatch
def copy_from_stdin_command(db: object, target_table: str,
                            csv_format: bool = None, skip_header: bool = None,
                            delimiter_char: str = None, quote_char: str = None,
                            null_value_string: str = None, timezone: str = None,
                            pipe_format: formats.Format = None) -> str:
    """
    Creates a shell command that receives data from stdin and writes it to a table.

    Options are tailored for the PostgreSQL `COPY FROM STDIN` command, adaptions might be needed for other databases.
    https://www.postgresql.org/docs/current/static/sql-copy.html

    Args:
        db: The database to use (either an alias or a `dbs.DB` object
        target_table: The table in which the data is written
        csv_format: Treat the input as a CSV file (comma separated, double quoted literals)
        skip_header: When true, skip the first line
        delimiter_char: The character that separates columns
        quote_char: The character for quoting strings
        null_value_string: The string that denotes NULL values
        timezone: Sets the timezone of the client, if applicable
        pipe_format: The format passed from stdin

    Returns:
        The composed shell command

    Example:
        >>> print(copy_from_stdin_command('mara', target_table='foo'))
        PGTZ=Europe/Berlin PGOPTIONS=--client-min-messages=warning psql --username=root --host=localhost \
            --echo-all --no-psqlrc --set ON_ERROR_STOP=on mara \
            --command="COPY foo FROM STDIN WITH CSV"
    """
    raise NotImplementedError(f'Please implement copy_from_stdin_command for type "{db.__class__.__name__}"')


@copy_from_stdin_command.register(str)
def __(alias: str, target_table: str, csv_format: bool = None, skip_header: bool = None,
       delimiter_char: str = None, quote_char: str = None, null_value_string: str = None,
       timezone: str = None, pipe_format: formats.Format = None):
    return copy_from_stdin_command(
        dbs.db(alias), target_table=target_table, csv_format=csv_format, skip_header=skip_header,
        delimiter_char=delimiter_char, quote_char=quote_char,
        null_value_string=null_value_string, timezone=timezone, pipe_format=pipe_format)


@copy_from_stdin_command.register(dbs.PostgreSQLDB)
def __(db: dbs.PostgreSQLDB, target_table: str, csv_format: bool = None, skip_header: bool = None,
       delimiter_char: str = None, quote_char: str = None, null_value_string: str = None, timezone: str = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                 quote_char=quote_char, null_value_string=null_value_string)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                            quote_char=quote_char, null_value_string=null_value_string)

    columns = ''
    sed_stdin = ''
    sql = f'COPY {target_table}{columns} FROM STDIN WITH'
    if isinstance(pipe_format, formats.JsonlFormat):
        columns = ' (' + ', '.join(['data']) + ')'
        # escapes JSON escapings since PostgreSQL interprets C-escapes in TEXT mode
        nl = '\n'
        sed_stdin += fr"sed 's/\\/\\x5C/g' \{nl}  | "

    elif isinstance(pipe_format, formats.CsvFormat):
        sql += ' CSV'
        if pipe_format.header:
            sql += ' HEADER'
        if pipe_format.delimiter_char is not None:
            sql += f" DELIMITER AS '{pipe_format.delimiter_char}'"
        if pipe_format.null_value_string is not None:
            sql += f" NULL AS '{pipe_format.null_value_string}'"
        if pipe_format.quote_char is not None:
            sql += f" QUOTE AS '{pipe_format.quote_char}'"

    elif isinstance(pipe_format, formats.NativeFormat):
        pass

    else:
        raise ValueError(f'Unsupported pipe_format for PostgreSQLDB: {pipe_format}')

    # escape double quotes
    sql = sql.replace('"', '\\"')

    return f'{sed_stdin}{query_command(db, timezone)} \\\n      --command="{sql}"'


@copy_from_stdin_command.register(dbs.RedshiftDB)
def __(db: dbs.RedshiftDB, target_table: str, csv_format: bool = None, skip_header: bool = None,
       delimiter_char: str = None, quote_char: str = None, null_value_string: str = None, timezone: str = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                 quote_char=quote_char, null_value_string=null_value_string)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                            quote_char=quote_char, null_value_string=null_value_string)

    import uuid
    import datetime

    tmp_file_name = f'tmp-{datetime.datetime.now().isoformat()}-{uuid.uuid4().hex}.csv'
    s3_write_command = f'AWS_ACCESS_KEY_ID={db.aws_access_key_id} AWS_SECRET_ACCESS_KEY={db.aws_secret_access_key} aws s3 cp - s3://{db.aws_s3_bucket_name}/{tmp_file_name}'
    s3_delete_tmp_file_command = f'AWS_ACCESS_KEY_ID={db.aws_access_key_id} AWS_SECRET_ACCESS_KEY={db.aws_secret_access_key} aws s3 rm s3://{db.aws_s3_bucket_name}/{tmp_file_name}'

    columns = ''
    sed_stdin = ''
    sql = f"COPY {target_table}{columns} FROM 's3://{db.aws_s3_bucket_name}/{tmp_file_name}' access_key_id '{db.aws_access_key_id}' secret_access_key '{db.aws_secret_access_key}'"
    if isinstance(pipe_format, formats.JsonlFormat):
        columns = ' (' + ', '.join(['data']) + ')'
        # escapes JSON escapings since PostgreSQL interprets C-escapes in TEXT mode
        nl = '\n'
        sed_stdin += fr"sed 's/\\/\\x5C/g' \{nl}| "

    elif isinstance(pipe_format, formats.CsvFormat):
        csv_format = True
        if delimiter_char is None:
            delimiter_char = pipe_format.delimiter_char
        if quote_char is None:
            quote_char = pipe_format.quote_char
        if skip_header is None:
            skip_header = pipe_format.header
        sql += ' CSV'
        if pipe_format.header:
            sql += ' HEADER'
        if pipe_format.delimiter_char is not None:
            sql += f" DELIMITER AS '{pipe_format.delimiter_char}'"
        if pipe_format.null_value_string is not None:
            sql += f" NULL AS '{pipe_format.null_value_string}'"
        if pipe_format.quote_char is not None:
            sql += f" QUOTE AS '{pipe_format.quote_char}'"

    elif isinstance(pipe_format, formats.NativeFormat):
        pass

    else:
        raise ValueError(f'Unsupported pipe_format for RedshiftDB: {pipe_format}')

    return s3_write_command + ' &&\n\n' \
            + f'{sed_stdin}{query_command(db, timezone)} \\\n      --command="{sql}" \\\n  || /bin/false \\\n  ; RC=$?\n\n' \
            + s3_delete_tmp_file_command+' &&\n  $(exit $RC) || /bin/false'


@copy_from_stdin_command.register(dbs.BigQueryDB)
def __(db: dbs.BigQueryDB, target_table: str, csv_format: bool = None, skip_header: bool = None,
       delimiter_char: str = None, quote_char: str = None, null_value_string: str = None, timezone: str = None,
       pipe_format: formats.Format = None):
    assert db.gcloud_gcs_bucket_name, f"Please provide the 'gcloud_gcs_bucket_name' parameter to database '{db}' "

    _check_format_with_args_used(pipe_format, header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                 quote_char=quote_char, null_value_string=null_value_string)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                            quote_char=quote_char, null_value_string=null_value_string)

    import uuid
    import datetime
    from .bigquery import bigquery_credentials

    service_account_email = bigquery_credentials(db).service_account_email

    set_env_prefix = f'CLOUDSDK_CORE_ACCOUNT={service_account_email}'
    bq_load_command = (set_env_prefix
                       + ' bq load'
                       + ' --headless'
                       + ' --quiet'
                       + (f' --location={db.location}' if db.location else '')
                       + (f' --project_id={db.project}' if db.project else '')
                       + (f' --dataset_id={db.dataset}' if db.dataset else '')
                       + (' --skip_leading_rows=1' if skip_header else '')
                       )

    if isinstance(pipe_format, formats.CsvFormat):
        bq_load_command += ' --source_format=CSV'
        file_extension = 'csv'

        if pipe_format.delimiter_char is not None:
            bq_load_command += f" --field_delimiter='{pipe_format.delimiter_char}'"
        if pipe_format.null_value_string is not None:
            bq_load_command += f" --null_marker='{pipe_format.null_value_string}'"
        if pipe_format.quote_char is not None:
            bq_load_command += f" --quote='{pipe_format.quote_char}'"

    elif isinstance(pipe_format, formats.JsonlFormat) or isinstance(pipe_format, formats.NativeFormat):
        bq_load_command += ' --source_format=NEWLINE_DELIMITED_JSON'
        file_extension = 'jsonl'

    elif isinstance(pipe_format, formats.AvroFormat):
        bq_load_command += ' --source_format=AVRO'
        file_extension = 'avro'

    elif isinstance(pipe_format, formats.ParquetFormat):
        bq_load_command += ' --source_format=PARQUET'
        file_extension = 'parquet'

    elif isinstance(pipe_format, formats.OrcFormat):
        bq_load_command += ' --source_format=ORC'
        file_extension = 'orc'

    else:
        raise ValueError(f'Unsupported pipe_format for BigQueryDB: {pipe_format}')

    tmp_file_name = f'tmp-{datetime.datetime.now().isoformat()}-{uuid.uuid4().hex}.{file_extension}'
    bq_load_command += f" '{target_table}'  gs://{db.gcloud_gcs_bucket_name}/{tmp_file_name}"

    gcs_write_command = f'{set_env_prefix} gsutil -q cp - gs://{db.gcloud_gcs_bucket_name}/{tmp_file_name}'
    gcs_delete_temp_file_command = f'{set_env_prefix} gsutil -q rm gs://{db.gcloud_gcs_bucket_name}/{tmp_file_name}'

    return gcs_write_command + '\\\n  \\\n  && ' \
           + bq_load_command + '\\\n  \\\n  && ' \
           + gcs_delete_temp_file_command


@copy_from_stdin_command.register(dbs.SqlcmdSQLServerDB)
def __(db: dbs.SqlcmdSQLServerDB, target_table: str, csv_format: bool = None, skip_header: bool = None,
       delimiter_char: str = None, quote_char: str = None, null_value_string: str = None, timezone: str = None,
       pipe_format: formats.Format = None):
    _check_format_with_args_used(pipe_format, header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                 quote_char=quote_char, null_value_string=null_value_string)
    if not pipe_format:
        pipe_format = _get_format_from_args(header=skip_header, delimiter_char=delimiter_char, csv_format=csv_format,
                                            quote_char=quote_char, null_value_string=null_value_string)

    if isinstance(pipe_format, formats.CsvFormat):
        if pipe_format.quote_char:
            raise ValueError('pipe_format.quote_char is not supported for SqlcmdSQLServerDB')
    else:
        raise ValueError(f'Unsupported pipe_format for SqlcmdSQLServerDB: {pipe_format}')

    if pipe_format.null_value_string is not None and pipe_format.null_value_string != '':
        raise ValueError("The parameter null_value_string must be None or an empty string ('') " + \
            "when the db_alias referres to a SQL Server (SqlcmdSQLServerDB)")

    if db.host:
        # connection to DB, see: https://docs.microsoft.com/en-us/sql/ssms/scripting/sqlcmd-connect-to-the-database-engine?view=sql-server-ver15
        if db.protocol == 'tcp':
            port = db.port if db.port else 1433
            server = f'tcp:{db.host},{port}'
        elif db.protocol == 'np':
            pipe = f'MSSQL${db.instance}\\sql\\query' if db.instance else 'pipe\\sql\\query'
            server = f'np:\\\\{db.host}\\{pipe}'
        elif db.protocol == 'lpc':
            server = f'lcp:{db.host}\\{db.instance}' if db.instance else f'lcp:{db.host}'
        else:
            if db.instance:
                server = f'{db.host}\\{db.instance}'
            elif db.port:
                server = f'{db.host},{db.port}'
            else:
                server = db.host

    return ('{ '
            # create a temporary file for stdin; bcp does not support stdin by default
            + 'TEMP_STDIN="$(mktemp)"; '
            # transforms CRLF to LF and saves stdin; bcp uses LF in unix systems by default
            + 'cat - | sed "s/\\r$//g" > "${TEMP_STDIN}"; '
            + f'bcp {target_table} in "${{TEMP_STDIN}}"'
            + (f' -U {db.user}' if db.user else '')
            + (f' -P {db.password}' if db.password else '')
            + (f' -S {server}' if server else '')
            + (' -u' if db.trust_server_certificate else '')
            + (f' -d {db.database}' if db.database else '')
            + ' -c'
            + (f' -t {pipe_format.delimiter_char or ","}' if pipe_format.delimiter_char != '\t' else '')
            + (' -F2' if pipe_format.header else '')
            # removes the temporary file
            + f'; rm -f "${{TEMP_STDIN}}" > /dev/null; '
            + '}')


# -------------------------------


@multidispatch
def copy_command(source_db: object, target_db: object, target_table: str,
                 timezone=None, csv_format=None, delimiter_char=None,
                 pipe_format: formats.Format = None) -> str:
    """
    Creates a shell command that
    - receives a sql query from stdin
    - executes the query in `source_db`
    - writes the results of the query to `target_table` in `target_db`

    Args:
        source_db: The database in which to run the query (either an alias or a `dbs.DB` object
        target_db: The database where to write the query results (alias or db configuration)
        target_table: The table in which to write the query results
        timezone: Sets the timezone of the client, if applicable
        csv_format: double quote 'difficult' strings
        delimiter_char: The character that separates columns, default '\t'
        pipe_format: The piping data format to be used


    Returns:
        A shell command string

    Example:
        >>> print(copy_command(dbs.SQLServerDB(database='source_db'), dbs.PostgreSQLDB(database='target_db'), \
                                'target_table'))
        sqsh  -D source_db -m csv \
          | PGTZ=Europe/Berlin PGOPTIONS=--client-min-messages=warning psql --echo-all --no-psqlrc \
               --set ON_ERROR_STOP=on target_db \
               --command="COPY target_table FROM STDIN WITH CSV HEADER"
    """
    raise NotImplementedError(
        f'Please implement copy_command for types "{source_db.__class__.__name__}" and "{target_db.__class__.__name__}"'
    )


@copy_command.register(str, str)
def __(source_db_alias: str, target_db_alias: str, target_table: str, timezone: str = None,
       csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    return copy_command(dbs.db(source_db_alias), dbs.db(target_db_alias),
                        target_table=target_table, timezone=timezone,
                        csv_format=csv_format, delimiter_char=delimiter_char,
                        pipe_format=pipe_format)


@copy_command.register(dbs.DB, str)
def __(source_db: dbs.DB, target_db_alias: str, target_table: str, timezone: str = None,
       csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    return copy_command(source_db, dbs.db(target_db_alias),
                        target_table=target_table, timezone=timezone,
                        csv_format=csv_format, delimiter_char=delimiter_char,
                        pipe_format=pipe_format)


@copy_command.register(dbs.PostgreSQLDB, dbs.PostgreSQLDB)
def __(source_db: dbs.PostgreSQLDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    return (copy_to_stdout_command(source_db, delimiter_char=delimiter_char, csv_format=csv_format,
                                   pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               null_value_string='', timezone=timezone, csv_format=csv_format,
                                               delimiter_char=delimiter_char, pipe_format=pipe_format))


@copy_command.register(dbs.PostgreSQLDB, dbs.BigQueryDB)
def __(source_db: dbs.PostgreSQLDB, target_db: dbs.BigQueryDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(
            delimiter_char='\t' if not delimiter_char and csv_format else delimiter_char)
    return (copy_to_stdout_command(source_db, delimiter_char=delimiter_char, csv_format=csv_format,
                                   pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               timezone=timezone, csv_format=csv_format,
                                               delimiter_char='\t' if not delimiter_char and csv_format else delimiter_char))


@copy_command.register(dbs.BigQueryDB, dbs.PostgreSQLDB)
def __(source_db: dbs.BigQueryDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(
            delimiter_char='\t' if not delimiter_char and csv_format else delimiter_char)
    return (copy_to_stdout_command(source_db, delimiter_char=delimiter_char, csv_format=csv_format,
                                   pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               timezone=timezone, csv_format=csv_format,
                                               pipe_format=pipe_format))


@copy_command.register(dbs.MysqlDB, dbs.PostgreSQLDB)
def __(source_db: dbs.MysqlDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               null_value_string='NULL', timezone=timezone,
                                               csv_format=csv_format, delimiter_char=delimiter_char,
                                               pipe_format=pipe_format))


@copy_command.register(dbs.MysqlDB, dbs.BigQueryDB)
def __(source_db: dbs.MysqlDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char)
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               null_value_string='NULL', timezone=timezone,
                                               csv_format=csv_format,
                                               pipe_format=pipe_format))


@copy_command.register(dbs.SQLServerDB, dbs.PostgreSQLDB)
def __(source_db: dbs.SQLServerDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char, header=True)
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table, csv_format=csv_format,
                                               skip_header=True, timezone=timezone,
                                               pipe_format=pipe_format))


@copy_command.register(dbs.SqshSQLServerDB, dbs.BigQueryDB)
def __(source_db: dbs.SQLServerDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char, header=True)
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table, csv_format=csv_format,
                                               skip_header=True, timezone=timezone,
                                               pipe_format=pipe_format))


@copy_command.register(dbs.SqlcmdSQLServerDB, dbs.BigQueryDB)
def __(source_db: dbs.SqlcmdSQLServerDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None):
    if csv_format is None:
        csv_format = True
    return (copy_to_stdout_command(source_db) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table, csv_format=csv_format,
                                               delimiter_char=delimiter_char,
                                               null_value_string='NULL', skip_header=True))


@copy_command.register(dbs.OracleDB, dbs.PostgreSQLDB)
def __(source_db: dbs.OracleDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char, header=False)
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               csv_format=csv_format, skip_header=False,
                                               null_value_string='NULL', timezone=timezone, pipe_format=pipe_format))


@copy_command.register(dbs.OracleDB, dbs.BigQueryDB)
def __(source_db: dbs.OracleDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char, header=False)
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table,
                                               csv_format=csv_format, skip_header=False,
                                               null_value_string='NULL', timezone=timezone, pipe_format=pipe_format))


@copy_command.register(dbs.SQLiteDB, dbs.PostgreSQLDB)
def __(source_db: dbs.SQLiteDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char, quote_char="''")
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table, timezone=timezone,
                                               null_value_string='NULL', csv_format=csv_format,
                                               pipe_format=pipe_format))


@copy_command.register(dbs.SQLiteDB, dbs.BigQueryDB)
def __(source_db: dbs.SQLiteDB, target_db: dbs.PostgreSQLDB, target_table: str,
       timezone: str = None, csv_format: bool = None, delimiter_char: str = None, pipe_format: formats.Format = None):
    if csv_format is None and pipe_format is None:
        pipe_format = formats.CsvFormat(delimiter_char=delimiter_char, quote_char="''")
    return (copy_to_stdout_command(source_db, pipe_format=pipe_format) + ' \\\n'
            + '  | ' + copy_from_stdin_command(target_db, target_table=target_table, timezone=timezone,
                                               null_value_string='NULL', quote_char="''", csv_format=csv_format,
                                               pipe_format=pipe_format))



"""Easy access to BigQuery databases via google.cloud.bigquery"""

import typing
from warnings import warn

import mara_db.dbs
import sys
import time
from google.api_core.exceptions import BadRequest


def bigquery_credentials(db: typing.Union[str, mara_db.dbs.BigQueryDB]) -> 'google.oauth2.service_account.Credentials':
    """Get the parsed service account"""
    from google.oauth2.service_account import Credentials

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    return Credentials.from_service_account_file(db.service_account_json_file_name)


def bigquery_client(db: typing.Union[str, mara_db.dbs.BigQueryDB]) -> 'google.cloud.bigquery.client.Client':
    """Get an bigquery client for a bq database alias"""
    from google.cloud.bigquery.client import Client

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    credentials = bigquery_credentials(db)

    return Client(project=credentials.project_id, credentials=credentials, location=db.location)


def bigquery_cursor_context(db: typing.Union[str, mara_db.dbs.BigQueryDB]) \
        -> 'google.cloud.bigquery.dbapi.cursor.Cursor':
    """Creates a context with a bigquery cursor for a database alias"""
    warn('Function bigquery_cursor_context(db) is deprecated. Please use db.cursor_context() instead.')

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    assert (isinstance(db, mara_db.dbs.BigQueryDB))

    return mara_db.dbs.cursor_context(db)


def create_bigquery_table_from_postgresql_query(
        postgresql_query: str, postgresql_db_alias: str,
        bigquery_db_alias: str, bigquery_dataset_id: str, bigquery_table_name: str):
    """
    Creates a table for bigquery from a Postgresql SELECT query. Will print the query

    Useful for copying PostgreSQL tables to BigQuery (create table first and then copy)

    Example:
        >>> create_bigquery_table_from_postgresql_query(
        >>>              postgresql_db_alias='dwh',
        >>>              postgresql_query='SELECT 1::SMALLINT AS a, now() as b',
        >>>              bigquery_db_alias='reporting',
        >>>              bigquery_dataset_id='foo',
        >>>              bigquery_table_name='bar')
        CREATE OR REPLACE TABLE `foo`.`bar` (
            `a` INT64,
            `b` TIMESTAMP
        )

    Args:
        postgresql_query: The query to execute in PostgreSQL, must not end with a semicolon
        postgresql_db_alias: The postgresql database to execute the query in
        bigquery_db_alias: The mara db alias of the bigquery connection
        bigquery_dataset_id: The id of the bigquery dataset in which the table is to be created
        bigquery_table_name: The name of the to be created table
    """
    from mara_db.postgresql import postgres_cursor_context
    with mara_db.postgresql.postgres_cursor_context(postgresql_db_alias) as cursor:
        cursor.execute('SELECT oid, typname FROM pg_type;')
        pg_types = {}
        for oid, type_name in cursor.fetchall():
            pg_types[oid] = type_name

        # https://cloud.google.com/bigquery/docs/reference/standard-sql/federated_query_functions#postgressql_mapping
        pg_to_bigquery_type_mapping = {
            'bool': 'BOOL',
            'bytea': 'BYTES',
            'date': 'DATE',
            'int2': 'INT64',
            'int4': 'INT64',
            'int8': 'INT64',
            'json': 'STRING',
            'jsonb': 'STRING',
            'numeric': 'NUMERIC',
            'float4': 'FLOAT64',
            'float8': 'FLOAT64',
            'varchar': 'STRING',
            'text': 'STRING',
            'time': 'TIME',
            'timestamp': 'DATETIME',
            'timestamptz': 'TIMESTAMP',
        }

        cursor.execute(postgresql_query + ' LIMIT 0')

        column_specs = []
        for column in cursor.description:
            pg_type = pg_types[column.type_code]
            assert pg_type in pg_to_bigquery_type_mapping, f"Unmapped type '{pg_type}'"
            column_specs.append(f'`{column.name}` {pg_to_bigquery_type_mapping[pg_type]}')

        query = f"""
CREATE OR REPLACE TABLE `{bigquery_dataset_id}`.`{bigquery_table_name}` (
    """ + ',\n    '.join(column_specs) + "\n)"

        print(query)

        client = bigquery_client(bigquery_db_alias)
        client.query(query)


def replace_dataset(db_alias: str, dataset_id: str, next_dataset_id: str):
    """
    Replaces the a bigquery dataset with the contents of another one.

    Args:
        db_alias: the mara db alias of the bigquery connection
        dataset_id: the dataset that will be replaced
        nextdata_set_id: the contents of the new dataset
    """
    print(f'replacing dataset `{dataset_id}` with contents of `{next_dataset_id}`')
    from mara_db.bigquery import bigquery_client

    client = bigquery_client(db_alias)

    # create target dataset if not exists
    client.create_dataset(dataset=dataset_id, exists_ok=True)

    # all tables in the next dataset
    next_tables = set([table.table_id for table in client.list_tables(next_dataset_id)])

    ddl = '\n'

    # delete tables in target dataset that are not in next dataset
    for table in client.list_tables(dataset_id):
        if table.table_id not in next_tables:
            ddl += f'DROP TABLE `{dataset_id}`.`{table.table_id}`; \n'

    # hopefully atomic operation
    for table_id in next_tables:
        ddl += f'CREATE OR REPLACE TABLE `{dataset_id}`.`{table_id}` AS SELECT * FROM `{next_dataset_id}`.`{table_id}`;\n'
        ddl += f'DROP TABLE `{next_dataset_id}`.`{table_id}`;\n'

    print(ddl)
    client.query(ddl)

    print(f'deleting dataset {next_dataset_id}')
    retries = 1
    # for some reason the 'DROP TABLE ...'  statements take some time, retry the data set deletion
    while True:
        try:
            client.delete_dataset(next_dataset_id)
            return
        except BadRequest as e:
            if retries <= 10:
                print(e, file=sys.stderr)
                seconds_to_sleep = retries * 4
                print(f'Waiting {seconds_to_sleep} seconds')
                time.sleep(seconds_to_sleep)
                retries += 1
            else:
                raise e



"""Different formats for piping"""
from typing import Optional


class Format:
    """Base format definition"""

    def __repr__(self) -> str:
        return (f'<{self.__class__.__name__}: '
                + ', '.join([f'{var}={getattr(self, var)}'
                             for var in vars(self) if getattr(self, var)])
                + '>')


class NativeFormat(Format):
    """Use the native format of e.g. a database."""
    def __init__(self):
        pass


class CsvFormat(Format):
    """
    CSV file format. See https://tools.ietf.org/html/rfc4180
    """
    def __init__(self, delimiter_char: str = ',', quote_char: Optional[str] = None, header: bool = False, footer: bool = False, null_value_string: Optional[str] = None):
        """
        CSV file format. See https://tools.ietf.org/html/rfc4180

        Args:
            delimiter_char: The character that separates columns
            quote_char: The character for quoting strings
            header: Whether a csv header with the column name(s) is part of the CSV file.
            footer: Whether a footer will be included or not. False by default.
            null_value_string: The string used to indicate NULL.
        """
        self.delimiter_char = delimiter_char or ','
        self.quote_char = quote_char
        self.header = header or False
        self.footer = footer or False
        self.null_value_string = null_value_string


class JsonlFormat(Format):
    """New line delimited JSON stream. See https://en.wikipedia.org/wiki/JSON_streaming"""
    def __init__(self):
        pass


class AvroFormat(Format):
    """Apache Avro"""
    def __init__(self):
        pass


class ParquetFormat(Format):
    """Apache Parquet"""
    def __init__(self):
        pass


class OrcFormat(Format):
    """Apache ORC"""
    def __init__(self):
        pass


def _check_format_with_args_used(pipe_format: Format, header: Optional[bool] = None, footer: Optional[bool] = None, delimiter_char: Optional[str] = None,
                                 csv_format: Optional[bool] = None, quote_char: Optional[str] = None, null_value_string: Optional[str] = None):
    if pipe_format:
        assert all(v is None for v in [header, footer, delimiter_char, csv_format, quote_char, null_value_string]), "You cannot pass format and an old parameter (header, footer, delimiter, csv_format, quote_char, null_value_string) at the same time"


def _get_format_from_args(header: Optional[bool] = None, footer: Optional[bool] = None, delimiter_char: Optional[str] = None, csv_format: Optional[bool] = None,
                          quote_char: Optional[str] = None, null_value_string: Optional[str] = None) -> Format:
    """A internal method handling old parameter settings"""
    if csv_format or delimiter_char and csv_format is None:
        return CsvFormat(delimiter_char=delimiter_char,
                         quote_char=quote_char,
                         header=header,
                         footer=footer,
                         null_value_string=null_value_string)
    else:
        return NativeFormat()



import functools

import sqlalchemy.engine
import sqlalchemy.sql.schema

import mara_db.dbs


@functools.singledispatch
def engine(db: object) -> sqlalchemy.engine.Engine:
    """
    Returns a sql alchemy engine for a configured database connection

    Args:
        db: The database to use (either an alias or a `dbs.DB` object

    Returns:
        The generated sqlalchemy engine

    Example:
        >>> print(engine('mara'))
        Engine(postgresql+psycopg2://None@localhost/mara)
    """
    pass


@engine.register(str)
def __(alias: str, **_):
    return engine(mara_db.dbs.db(alias))


@engine.register(mara_db.dbs.DB)
def __(db: mara_db.dbs.DB, **_):
    return sqlalchemy.create_engine(db.sqlalchemy_url)


@engine.register(mara_db.dbs.BigQueryDB)
def __(db: mara_db.dbs.BigQueryDB):
    # creates bigquery dialect
    url = db.sqlalchemy_url

    return sqlalchemy.create_engine(url,
                                    credentials_path=db.service_account_json_file_name,
                                    location=db.location)


@engine.register(mara_db.dbs.DatabricksDB)
def __(db: mara_db.dbs.DatabricksDB):
    url = db.sqlalchemy_url

    return sqlalchemy.create_engine(url,
                                    connect_args={
                                        "http_path": db.http_path
                                    })



import pytest
import subprocess

from mara_db import shell, sqlalchemy_engine
from tests.local_config import SNOWFLAKE_DB


if not SNOWFLAKE_DB:
    pytest.skip("skipping SnowflakeDB tests: variable SNOWFLAKE_DB not set", allow_module_level=True)


def test_snowflake_query_command():
    command = 'echo "SELECT 1" \\\n'
    command += '  | ' + shell.query_command(SNOWFLAKE_DB)
    assert command

    print(command)
    (exitcode, _) = subprocess.getstatusoutput(command)
    assert exitcode == 0


def test_snowflake_copy_to_stdout():
    command = 'echo "SELECT 1 AS Col1, \'FOO\' AS Col2 UNION ALL SELECT 2, \'BAR\'" \\\n'
    command += '  | ' + shell.copy_to_stdout_command(SNOWFLAKE_DB,
        csv_format=True,
        header=True,
        delimiter_char=',')
    assert command

    print(command)
    (exitcode, pstdout) = subprocess.getstatusoutput(command)
    assert exitcode == 0
    print(pstdout)
    assert pstdout == '''"COL1","COL2"
"1","FOO"
"2","BAR"'''


def test_snowflake_sqlalchemy():
    from sqlalchemy import text
    engine = sqlalchemy_engine.engine(SNOWFLAKE_DB)
    with engine.connect() as con:
        con.execute(statement = text("SELECT 1"))



"""Configuration of database connections"""
import typing

from mara_db import dbs


def databases() -> typing.Dict[str, dbs.DB]:
    """The list of database connections to use, by alias"""
    return {}


def default_timezone() -> str:
    """
    The default timezone to be used for database connections
    See: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """
    return 'Europe/Berlin'


def default_echo_queries() -> bool:
    """
    If queries should be printed on execution by default, if applicable
    """
    return True


def schema_ui_foreign_key_column_regex() -> typing.Pattern:
    """A regex that classifies a table column as being used in a foreign constraint (for coloring missing constraints)"""
    return r'.*_fk$'



"""Easy access to Databricks databases via databricks-sql-connector"""

import typing
from warnings import warn

import mara_db.dbs


def databricks_cursor_context(db: typing.Union[str, mara_db.dbs.DatabricksDB]) \
        -> 'databricks.sql.client.Cursor':
    warn('Function databricks_cursor_context(db) is deprecated. Please use db.cursor_context() instead.')

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    assert (isinstance(db, mara_db.dbs.DatabricksDB))

    return mara_db.dbs.cursor_context(db)



"""Easy access to MySQL databases via mysql-client"""

import typing
from warnings import warn

import mara_db.dbs


def mysql_cursor_context(db: typing.Union[str, mara_db.dbs.MysqlDB]) -> 'MySQLdb.cursors.Cursor':
    """Creates a context with a mysql-client cursor for a database alias or database"""
    warn('Function mysql_cursor_context(db) is deprecated. Please use db.cursor_context() instead.')

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    assert (isinstance(db, mara_db.dbs.MysqlDB))

    return mara_db.dbs.cursor_context(db)



import pytest
import subprocess

from mara_db import shell, sqlalchemy_engine
from tests.local_config import DATABRICKS_DB


if not DATABRICKS_DB:
    pytest.skip("skipping DatabricksDB tests: variable DATABRICKS_DB not set", allow_module_level=True)


def test_databricks_query_command():
    command = 'echo "SELECT 1" \\\n'
    command += '  | ' + shell.query_command(DATABRICKS_DB)
    assert command

    print(command)
    (exitcode, _) = subprocess.getstatusoutput(command)
    assert exitcode == 0


def test_databricks_copy_to_stdout():
    command = 'echo "SELECT 1 AS Col1, \'FOO\' AS Col2 UNION ALL SELECT 2, \'BAR\'" \\\n'
    command += '  | ' + shell.copy_to_stdout_command(DATABRICKS_DB,
        csv_format=True,
        header=True,
        delimiter_char=',')
    assert command

    print(command)
    (exitcode, pstdout) = subprocess.getstatusoutput(command)
    assert exitcode == 0
    print(pstdout)
    assert pstdout == '''Col1,Col2
1,FOO
2,BAR'''


def test_databricks_sqlalchemy():
    from sqlalchemy import text
    engine = sqlalchemy_engine.engine(DATABRICKS_DB)
    with engine.connect() as con:
        con.execute(statement = text("SELECT 1"))


def test_databricks_connect():
    """
    A simple test to check if the connect API works.
    """
    from .db_test_helper import _test_connect
    _test_connect(DATABRICKS_DB)


def test_databricks_cursor_context():
    """
    A simple test to check if the cursor context of the db works.
    """
    from .db_test_helper import _test_cursor_context
    _test_cursor_context(DATABRICKS_DB)



"""Easy access to SQLServer databases via pyodbc-client"""

import typing
from warnings import warn

import mara_db.dbs


def sqlserver_cursor_context(db: typing.Union[str, mara_db.dbs.SQLServerDB]) -> 'pyodbc.Cursor':
    """Creates a context with a pyodbc-client cursor for a database alias or database"""
    warn('Function sqlserver_cursor_context(db) is deprecated. Please use db.cursor_context() instead.')

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    assert (isinstance(db, mara_db.dbs.SQLServerDB))

    return mara_db.dbs.cursor_context(db)



import pytest

def test_my():
    import mara_db.dbs

    db = mara_db.dbs.SqlcmdSQLServerDB(host='ABC123', user='A', password="a", trust_server_certificate=True)

    odbc = db.sqlalchemy_url

    print(odbc)

    assert False



"""DB schema visualization"""

import datetime
import re
import typing
from functools import singledispatch
from html import escape

import flask
from mara_db import config, dbs
from mara_page import acl, navigation, response, bootstrap, html, _, xml

blueprint = flask.Blueprint('mara_db', __name__, static_folder='static', template_folder='templates', url_prefix='/db')

acl_resource = acl.AclResource(name='DB Schema')


def navigation_entry():
    return navigation.NavigationEntry(
        label='DB Schema', icon='star', description='Schemas of all databases connections',
        children=[navigation.NavigationEntry(
            label='Overview', icon='list',
            uri_fn=lambda: flask.url_for('mara_db.index_page'))] +
                 [
                     navigation.NavigationEntry(
                         label=alias, icon='database',
                         description=f'The schema of the {alias} db',
                         uri_fn=lambda current_db=alias: flask.url_for('mara_db.schema_page', db_alias=current_db))
                     for alias, db in config.databases().items()
                     if supports_extract_schema(db)
                 ])


@blueprint.route('/')
def index_page():
    """Overview page of mara_db"""
    return response.Response(
        title=f'Database schemas',
        html=bootstrap.card(
            body=[_.div(style='display:inline-block; margin-top:15px; margin-bottom:15px; margin-right:50px;')[
                      _.a(href=flask.url_for('mara_db.schema_page', db_alias=db_alias))[
                          _.span(class_='fa fa-database')[''], ' ', db_alias],
                      _.br,
                      _.span(style='color:#888')[escape(str(type(db).__name__))]
                  ]
                  for db_alias, db in config.databases().items()]),

        js_files=[flask.url_for('mara_db.static', filename='schema-page.js')])


@singledispatch
def supports_extract_schema(db: object) -> [bool]:
    """
    Returns true when the db supports schema extraction

    Args:
        db: The database which shall be tested for schema extraction
    """
    return False


@supports_extract_schema.register(str)
def __(alias: str):
    return supports_extract_schema(dbs.db(alias))


@supports_extract_schema.register(dbs.PostgreSQLDB)
def __(db: dbs.PostgreSQLDB):
    return True


@supports_extract_schema.register(dbs.RedshiftDB)
def __(db: dbs.RedshiftDB):
    return False


@supports_extract_schema.register(dbs.BigQueryDB)
def __(db: dbs.BigQueryDB):
    # BigQuery does not support primary and foreign key relations
    return False


@supports_extract_schema.register(dbs.MysqlDB)
def __(db: dbs.MysqlDB):
    return True


@supports_extract_schema.register(dbs.SQLServerDB)
def __(db: dbs.SQLServerDB):
    # check if module pyodbc can be imported
    import importlib
    pyodbc_spec = importlib.util.find_spec("pyodbc")
    return pyodbc_spec is not None


@blueprint.route('/<string:db_alias>')
def schema_page(db_alias: str):
    """A page that visiualizes the schemas of a database"""
    if db_alias not in config.databases():
        flask.abort(404, f'unkown database {db_alias}')

    return response.Response(
        title=f'Schema of database {db_alias}',
        html=[bootstrap.card(sections=[
            html.asynchronous_content(flask.url_for('mara_db.schema_selection', db_alias=db_alias)),
            [_.div(id='schema-container')]]),
            html.spinner_js_function()],
        js_files=[flask.url_for('mara_db.static', filename='schema-page.js')],
        action_buttons=[response.ActionButton(
            action='javascript:schemaPage.downloadSvg()', label='SVG',
            title='Save current chart as SVG file', icon='download')]
    )


@singledispatch
def schemas_with_foreign_key_constraints(db: object) -> [str]:
    """
    Returns all schemas that are effected by foreign key constraints

    Args:
        db: The database in which to run the query (either an alias or a `dbs.DB` object
    """
    raise NotImplementedError(
        f'Please implement schemas_with_foreign_key_constraints for type "{db.__class__.__name__}"')


@schemas_with_foreign_key_constraints.register(str)
def __(alias: str):
    return schemas_with_foreign_key_constraints(dbs.db(alias))


@schemas_with_foreign_key_constraints.register(dbs.PostgreSQLDB)
def __(db: dbs.PostgreSQLDB):
    import mara_db.postgresql
    with mara_db.postgresql.postgres_cursor_context(db) as cursor:
        cursor.execute('''
SELECT
  array_cat(array_agg(DISTINCT constrained_table_schema.nspname), array_agg(DISTINCT referenced_table_schema.nspname))
FROM pg_constraint
  JOIN pg_class constrained_table ON constrained_table.oid = pg_constraint.conrelid
  JOIN pg_namespace constrained_table_schema ON constrained_table.relnamespace = constrained_table_schema.oid
  JOIN pg_class referenced_table ON referenced_table.oid = pg_constraint.confrelid
  JOIN pg_namespace referenced_table_schema ON referenced_table.relnamespace = referenced_table_schema.oid''')
        result = cursor.fetchone()
        if result != (None,):
            return list(set(result[0]))


@schemas_with_foreign_key_constraints.register(dbs.MysqlDB)
def __(db: dbs.MysqlDB):
    import mara_db.mysql
    with mara_db.mysql.mysql_cursor_context(db) as cursor:
        cursor.execute("""
SELECT DISTINCT table_schema
FROM information_schema.table_constraints
WHERE CONSTRAINT_TYPE = 'FOREIGN KEY'
UNION
SELECT DISTINCT REFERENCED_TABLE_SCHEMA
FROM information_schema.key_column_usage
WHERE REFERENCED_TABLE_SCHEMA IS NOT NULL;
""")
        return [row[0] for row in cursor.fetchall()]


@schemas_with_foreign_key_constraints.register(dbs.SQLServerDB)
def __(db: dbs.SQLServerDB):
    import mara_db.sqlserver
    with mara_db.sqlserver.sqlserver_cursor_context(db) as cursor:
        cursor.execute("""
SELECT DISTINCT fpos.name AS schema_name
FROM sys.foreign_keys AS f
INNER JOIN sys.foreign_key_columns AS fc ON f.object_id = fc.constraint_object_id
INNER JOIN sys.objects AS fpo ON fpo.object_id = f.parent_object_id
INNER JOIN sys.schemas AS fpos ON fpos.schema_id = fpo.schema_id
""")
        return [row[0] for row in cursor.fetchall()]


@blueprint.route('/<string:db_alias>/.schemas')
def schema_selection(db_alias: str):
    """Asynchronously computes the list of schemas with foreign key constraints"""
    schemas_with_fk_constraints = schemas_with_foreign_key_constraints(db_alias)

    if not schemas_with_fk_constraints or len(schemas_with_fk_constraints) == 0:
        return str(_.i['No schemas with foreign key constraints found'])

    return ''.join(xml.render([
        [_.div(class_='form-check form-check-inline')[
             _.label(class_='form-check-label')[
                 _.input(class_="form-check-input schema-checkbox", type="checkbox", value=schema_name)[
                     ''], ' ', schema_name]]
         for schema_name in sorted(schemas_with_fk_constraints)],
        ' &#160;&#160;',
        _.div(class_='form-check form-check-inline')[
            _.label(class_='form-check-label')[
                _.input(class_="form-check-input", id='hide-columns-checkbox', type="checkbox")[
                    ''], ' ', 'hide columns']],
        ' &#160;&#160;',
        _.div(class_='form-check form-check-inline')[
            _.label(class_='form-check-label')[
                'graphviz engine ',
                _.select(id='engine', style='border:none;background-color:white;')[
                    [_.option(value=engine)[engine] for engine in ['neato', 'dot', 'twopi', 'fdp']]
                ]]],
        _.script['''
var schemaPage = SchemaPage("''' + flask.url_for('mara_db.schema_page', db_alias=db_alias) + '''", "''' + db_alias + '''");
''']]))


@singledispatch
def extract_schema(db: object, schema_names: [str]) -> (typing.Dict, typing.Set):
    """
    Extracts foreign key constraints and the involved tables from a db

    Args:
        db: The database in which to run the query (either an alias or a `dbs.DB` object
        schema_names: the schemas to visualize

    Returns:
        A dictionary of tables:
            {(table_schema, table_name): {'columns': [columns], 'constrained-columns': {constrained-columns}}

        All foreign key constrains as a set of tuples:
            {((table_schema, table_name), (referenced_schema_name, referenced_table_name))}
    """
    raise NotImplementedError(f'Please implement extract_schema for type "{db.__class__.__name__}"')


@extract_schema.register(str)
def __(alias: str, schema_names: [str]):
    return extract_schema(dbs.db(alias), schema_names)


@extract_schema.register(dbs.PostgreSQLDB)
def __(db: dbs.PostgreSQLDB, schema_names: [str]):
    import mara_db.postgresql

    # get all table inheritance relations as dictionary: {(child_schema, child_table): (parent_schema, parent_table)
    inherited_tables = {}
    with mara_db.postgresql.postgres_cursor_context(db) as cursor:
        cursor.execute("""
SELECT
  rel_namespace.nspname, rel.relname ,
  parent_namespace.nspname, parent.relname
FROM pg_inherits
  JOIN pg_class parent ON parent.oid = pg_inherits.inhparent
  JOIN pg_class rel ON rel.oid = pg_inherits.inhrelid
  JOIN pg_namespace parent_namespace ON parent_namespace.oid = parent.relnamespace
  JOIN pg_namespace rel_namespace ON rel_namespace.oid = rel.relnamespace""")
        for schema_name, table_name, parent_schema_name, parent_table_name in cursor.fetchall():
            inherited_tables[(schema_name, table_name)] = (parent_schema_name, parent_table_name)

    # get all tables that have foreign key constrains on them or are referenced by foreign key constraints
    tables = {}  # {(table_schema, table_name): {'columns': [columns], 'constrained-columns': {constrained-columns}}
    foreign_key_constraints = set()  # {((table_schema, table_name), (referenced_schema_name, referenced_table_name)}

    def empty_table():
        return {'columns': [], 'constrained-columns': set()}

    with mara_db.postgresql.postgres_cursor_context(db) as cursor:
        cursor.execute(f'''
SELECT
  constrained_table_schema.nspname,
  constrained_table.relname,
  array_agg(constrained_column.attname),
  referenced_table_schema.nspname,
  referenced_table.relname
FROM pg_constraint
  JOIN pg_class constrained_table ON constrained_table.oid = pg_constraint.conrelid
  JOIN pg_namespace constrained_table_schema ON constrained_table.relnamespace = constrained_table_schema.oid
  JOIN pg_class referenced_table ON referenced_table.oid = pg_constraint.confrelid
  JOIN pg_namespace referenced_table_schema ON referenced_table.relnamespace = referenced_table_schema.oid
  JOIN pg_attribute constrained_column ON constrained_column.attrelid = constrained_table.oid AND attnum = ANY (conkey)
WHERE constrained_table_schema.nspname = ANY ({'%s'})
GROUP BY constrained_table_schema.nspname, constrained_table.relname, referenced_table_schema.nspname, referenced_table.relname;
''', (schema_names,))
        for schema_name, table_name, table_columns, referenced_schema_name, referenced_table_name in cursor.fetchall():
            referring_table = (schema_name, table_name)
            if referring_table in inherited_tables:
                referring_table = inherited_tables[referring_table]
            if referring_table not in tables:
                tables[referring_table] = empty_table()
            tables[referring_table]['constrained-columns'].update(table_columns)

            referenced_table = (referenced_schema_name, referenced_table_name)
            if referenced_table in inherited_tables:
                referenced_table = inherited_tables[referenced_table]

            if referenced_table not in tables:
                tables[referenced_table] = empty_table()

            foreign_key_constraints.add((referring_table, referenced_table))

    # get enum usages
    with mara_db.postgresql.postgres_cursor_context(db) as cursor:
        cursor.execute(f'''
SELECT
  DISTINCT
  pg_namespace_table.nspname AS table_schema,
  pg_class_table.relname     AS table_name,

  pg_namespace_enum.nspname  AS enum_schema,
  pg_type.typname            AS enum_type
FROM pg_attribute
  JOIN pg_class pg_class_table ON pg_class_table.oid = attrelid
  JOIN pg_namespace pg_namespace_table ON pg_namespace_table.oid = pg_class_table.relnamespace
  JOIN pg_type ON atttypid = pg_type.OID
  JOIN pg_namespace pg_namespace_enum ON typnamespace = pg_namespace_enum.oid
  JOIN pg_enum ON pg_enum.enumtypid = pg_type.oid
WHERE pg_namespace_table.nspname = ANY ({'%s'})''', (schema_names,))
        for table_schema, table_name, enum_schema, enum_name in cursor.fetchall():
            if (table_schema, table_name) in tables:
                if not (enum_schema, enum_name) in tables:
                    tables[(enum_schema, enum_name)] = empty_table()

                foreign_key_constraints.add(((table_schema, table_name), (enum_schema, enum_name)))

    # get all columns of all tables
    with mara_db.postgresql.postgres_cursor_context(db) as cursor:
        cursor.execute('''
SELECT
  table_schema, table_name,
  array_agg(column_name :: TEXT ORDER BY ordinal_position)
FROM information_schema.columns
GROUP BY table_schema, table_name''')
        for schema_name, table_name, columns in cursor.fetchall():
            if (schema_name, table_name) in tables:
                tables[(schema_name, table_name)]['columns'] = columns

    return tables, foreign_key_constraints


@extract_schema.register(dbs.MysqlDB)
def __(db: dbs.MysqlDB, schema_names: [str]):
    import mara_db.mysql

    # get all tables that have foreign key constrains on them or are referenced by foreign key constraints
    tables = {}  # {(table_schema, table_name): {'columns': [columns], 'constrained-columns': {constrained-columns}}
    foreign_key_constraints = set()  # {((table_schema, table_name), (referenced_schema_name, referenced_table_name)}

    def empty_table():
        return {'columns': [], 'constrained-columns': set()}

    with mara_db.mysql.mysql_cursor_context(db) as cursor:
        cursor.execute(f'''
SELECT i.table_schema,
       i.table_name,
       k.column_name,
       k.referenced_table_schema,
       k.referenced_table_name
FROM information_schema.table_constraints i
         LEFT JOIN information_schema.KEY_COLUMN_USAGE k
                   ON i.constraint_name = k.constraint_name
WHERE i.constraint_type = 'FOREIGN KEY' 
      AND k.referenced_table_name IS NOT NULL
      AND i.table_schema IN {'%s'}; ''', (schema_names,))
        for table_schema, table_name, column_name, referenced_table_schema, referenced_table_name in cursor.fetchall():
            referring_table = (table_schema, table_name)
            referenced_table = (referenced_table_schema, referenced_table_name)
            if not referring_table in tables:
                tables[referring_table] = empty_table()
            tables[referring_table]['constrained-columns'].add(column_name)

            if not referenced_table in tables:
                tables[referenced_table] = empty_table()

            foreign_key_constraints.add((referring_table, referenced_table))

    with mara_db.mysql.mysql_cursor_context(db) as cursor:
        cursor.execute(f'''
SELECT table_schema, table_name, column_name
FROM information_schema.COLUMNS
WHERE table_schema IN {'%s'}
''', (schema_names,))
        for table_schema, table_name, column_name in cursor.fetchall():
            if (table_schema, table_name) in tables:
                tables[(table_schema, table_name)]['columns'].append(column_name)

    return tables, foreign_key_constraints


@extract_schema.register(dbs.SQLServerDB)
def __(db: dbs.SQLServerDB, schema_names: [str]):
    import mara_db.sqlserver

    # get all tables that have foreign key constrains on them or are referenced by foreign key constraints
    tables = {}  # {(table_schema, table_name): {'columns': [columns], 'constrained-columns': {constrained-columns}}
    foreign_key_constraints = set()  # {((table_schema, table_name), (referenced_schema_name, referenced_table_name)}

    def empty_table():
        return {'columns': [], 'constrained-columns': set()}

    with mara_db.sqlserver.sqlserver_cursor_context(db) as cursor:
        cursor.execute(f'''
SELECT
	s.name AS table_schema,
	t.name AS table_name,
	COL_NAME(fkc.parent_object_id, fkc.parent_column_id) AS column_name,
	fkts.name AS referenced_table_name,
	OBJECT_NAME (fk.referenced_object_id) AS referenced_table_name
FROM sys.tables t
INNER JOIN sys.schemas s ON
	s.schema_id = t.schema_id
LEFT JOIN sys.foreign_keys fk ON
	fk.parent_object_id = t.object_id
LEFT JOIN sys.foreign_key_columns fkc ON
	fkc.constraint_object_id = fk.object_id
LEFT JOIN sys.tables fkt ON
	fkt.object_id = fk.referenced_object_id
LEFT JOIN sys.schemas fkts ON
	fkts.schema_id = fkt.schema_id
WHERE s.name IN ('%s'); ''' % '\',\''.join(schema_names))
        for table_schema, table_name, column_name, referenced_table_schema, referenced_table_name in cursor.fetchall():
            referring_table = (table_schema, table_name)
            referenced_table = (referenced_table_schema, referenced_table_name)
            if not referring_table in tables:
                tables[referring_table] = empty_table()
            if column_name is not None:
                tables[referring_table]['constrained-columns'].add(column_name)

            # this logic is necessary so that tables with no foreign key are not added to the schema
            if referenced_table_schema is not None and referenced_table_name is not None:
                if not referenced_table in tables:
                    tables[referenced_table] = empty_table()
                foreign_key_constraints.add((referring_table, referenced_table))

    with mara_db.sqlserver.sqlserver_cursor_context(db) as cursor:
        cursor.execute(f'''
SELECT
	s.name AS table_schema,
	t.name AS table_name,
	c.name AS column_name
FROM sys.columns c
INNER JOIN sys.tables t ON
	t.object_id = c.object_id
INNER JOIN sys.schemas s ON
	s.schema_id = t.schema_id
WHERE s.name IN ('%s')
''' % '\',\''.join(schema_names))
        for table_schema, table_name, column_name in cursor.fetchall():
            if (table_schema, table_name) in tables:
                tables[(table_schema, table_name)]['columns'].append(column_name)
    return tables, foreign_key_constraints


@blueprint.route('/<string:db_alias>/<path:schemas>')
@acl.require_permission(acl_resource, do_abort=False)
def draw_schema(db_alias: str, schemas: str):
    """Shows a chart of the tables and FK relationships in a given database and schema list"""

    if db_alias not in config.databases():
        flask.abort(404, f'unkown database {db_alias}')

    if not supports_extract_schema(db_alias):
        flask.abort(404, f"could not extract schema for database {db_alias}")

    schema_names = schemas.split('/')
    hide_columns = flask.request.args.get('hide-columns')
    engine = flask.request.args.get('engine', 'neato')

    tables, fk_constraints = extract_schema(db_alias, schema_names)

    import graphviz.backend

    graph = graphviz.Digraph(engine=engine,
                             graph_attr={'splines': 'True', 'overlap': 'ortho'})

    schema_colors = {}
    fk_pattern = re.compile(config.schema_ui_foreign_key_column_regex())
    for schema_name, table_name in sorted(tables):
        if schema_name not in schema_colors:
            colors = ['#ffffcc', '#bbffcc', '#cceeff', '#eedd99', '#ddee99', '#99ddff', '#dddddd']
            schema_colors[schema_name] = colors[len(schema_colors) % len(colors)]

        label = '< <TABLE BORDER="1" CELLBORDER="0" CELLSPACING="0" CELLPADDING="1" BGCOLOR="' \
                + schema_colors[schema_name] + '"><TR>'

        node_name = schema_name + '.' + table_name
        if hide_columns:
            label += '<TD ALIGN="LEFT"> ' + table_name.replace('_', '<BR/>') + ' </TD></TR>'
        else:
            label += '<TD ALIGN="LEFT"><U><B> ' + table_name + ' </B></U></TD></TR>'
            for column in tables[(schema_name, table_name)]['columns']:
                label += '<TR><TD ALIGN="LEFT" > '
                if fk_pattern.match(column) \
                        and column not in tables[(schema_name, table_name)]['constrained-columns']:
                    label += '<B><I><FONT COLOR="#dd55dd"> ' + column + ' </FONT></I></B>'
                else:
                    label += column
                label += ' </TD></TR>'

        label += '</TABLE> >'

        graph.node(name=node_name, label=label,
                   _attributes={'fontname': 'Helvetica, Arial, sans-serif', 'fontsize': '10',
                                'fontcolor': '#555555', 'shape': 'none'})

    for (schema_name, table_name), (referenced_schema_name, referenced_table_name) in fk_constraints:
        graph.edge(schema_name + '.' + table_name, referenced_schema_name + '.' + referenced_table_name,
                   _attributes={'color': '#888888'})

    try:
        svg = graph.pipe('svg').decode('utf-8')
    except graphviz.backend.ExecutableNotFound as e:
        import uuid
        # This exception occurs when the graphviz tools are not found.
        # We use here a fallback to client-side rendering using the javascript library d3-graphviz.
        graph_id = f'dependency_graph_{uuid.uuid4().hex}'
        escaped_graph_source = graph.source.replace("`","\\`")
        return str(_.div(id=graph_id)[
            _.tt(style="color:red")[str(e)],
        ]) + str(_.script[
            f'div=d3.select("#{graph_id}");',
            'graph=div.graphviz();',
            'div.text("");',
            f'graph.renderDot(`{escaped_graph_source}`);',
        ])

    response = flask.Response(svg)
    response.headers[
        'Content-Disposition'] = f'attachment; filename="{datetime.date.today().isoformat()}-{db_alias}.svg"'
    return response



"""Abstract definition of database connections"""

import contextlib
import functools
import pathlib
from typing import Union


@functools.lru_cache(maxsize=None)
def db(alias):
    """Returns a database configuration by alias"""
    from . import config
    databases = config.databases()
    if alias not in databases:
        raise KeyError(f'database alias "{alias}" not configured')
    return databases[alias]


class DB:
    """Generic database connection definition"""

    def __repr__(self) -> str:
        return (f'<{self.__class__.__name__}: '
                + ', '.join([f'{var}={"*****" if (var == "password" or "secret" in var) else getattr(self, var)}'
                             for var in vars(self) if getattr(self, var)])
                + '>')

    @property
    def sqlalchemy_url(self):
        """Returns the SQLAlchemy url for a database"""
        raise NotImplementedError(f'Please implement sqlalchemy_url for type "{self.__class__.__name__}"')


class PostgreSQLDB(DB):
    def __init__(self, host: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None,
                 sslmode: str = None, sslrootcert: str = None, sslcert: str = None, sslkey: str = None):
        """
        Connection information for a PostgreSQL database

        For the SSL options, see https://www.postgresql.org/docs/current/libpq-ssl.html#LIBPQ-SSL-PROTECTION
        """
        self.host = host
        self.database = database
        self.port = port
        self.user = user
        self.password = password

        self.sslmode = sslmode
        self.sslrootcert = sslrootcert
        self.sslcert = sslcert
        self.sslkey = sslkey

    @property
    def sqlalchemy_url(self):
        return (f'postgresql+psycopg2://{self.user}{":" + self.password if self.password else ""}@{self.host}'
                + f'{":" + str(self.port) if self.port else ""}/{self.database}')


class RedshiftDB(PostgreSQLDB):
    def __init__(self, host: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None,
                 aws_access_key_id=None, aws_secret_access_key=None, aws_s3_bucket_name=None):
        """
        Connection information for a RedShift database

        The aws_* parameters are for copying to Redshift from stdin via an s3 bucket
        (requires the https://pypi.org/project/awscli/) package to be installed)
        """
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_s3_bucket_name = aws_s3_bucket_name
        super(RedshiftDB, self).__init__(host, port, database, user, password)


class BigQueryDB(DB):
    def __init__(self,
                 service_account_json_file_name: str,
                 location: str = None, project: str = None, dataset: str = None,
                 gcloud_gcs_bucket_name=None, use_legacy_sql: bool = False):
        """
        Connection information for a BigQueryDB database

        Enabling the BigQuery API and Service account json credentials are required. For more:
        https://cloud.google.com/bigquery/docs/quickstarts/quickstart-client-libraries#before-you-begin

        Args:
            service_account_json_file_name: The name of the private key file provided by Google when creating a service account (in json format)
            location: Default geographic location to use when creating datasets or determining where jobs should run
            project: Default project to use for requests.
            dataset: Default dataset to use for requests.
            gcloud_gcs_bucket_name: The Google Cloud Storage bucked used as cache for loading data
            use_legacy_sql: (default: false) If true, use the old BigQuery SQL dialect is used.
        """
        self.service_account_json_file_name = service_account_json_file_name
        self.location = location
        self.project = project
        self.dataset = dataset
        self.gcloud_gcs_bucket_name = gcloud_gcs_bucket_name
        self.use_legacy_sql = use_legacy_sql

    @property
    def sqlalchemy_url(self):
        url = 'bigquery://'
        if self.project:
            url += self.project
            if self.dataset:
                url += '/' + self.dataset
        return url


class MysqlDB(DB):
    def __init__(self, host: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None, ssl: bool = None, charset: str = None):
        self.host = host
        self.database = database
        self.port = port
        self.user = user
        self.password = password
        self.ssl = ssl
        self.charset = charset


class SQLServerDB(DB):
    def __new__(cls, host: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None, odbc_driver: str = None,
                 **kargs):
        """
        Connection information for a SQL Server database
        """
        if cls is SQLServerDB:
            # Here we define what happens when the class is directly created in code
            #
            # We defined here that class SqshSQLServerDB shall be used by default. In a newer
            # major version we could change this to SqlcmdSQLServerDB but we do not want to
            # introduce a breaking change here at this point.
            return SqshSQLServerDB(host=host, port=port, database=database, user=user, password=password, odbc_driver=odbc_driver)
        else:
            # This is called when the class is created from a derived class (e.g. SqshSQLServerDB)
            return super(SQLServerDB, cls).__new__(cls)

    def __init__(self, host: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None, odbc_driver: str = None):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password
        if odbc_driver is None:
            self.odbc_driver = 'ODBC Driver 17 for SQL Server' # default odbc driver
        else:
            self.odbc_driver = odbc_driver

    @property
    def sqlalchemy_url(self):
        import urllib.parse
        port = self.port if self.port else 1433
        driver = self.odbc_driver.replace(' ','+')
        return f'mssql+pyodbc://{self.user}:{urllib.parse.quote(self.password)}@{self.host}:{port}/{self.database}?driver={driver}'


class SqshSQLServerDB(SQLServerDB):
    def __init__(self, host: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None, odbc_driver: str = None):
        """
        Connection information for a SQL Server database using the unix package sqsh
        """
        # NOTE: The support for named instances is not added because the command `sqsh` does not support it
        super().__init__(host=host, port=port, database=database, user=user, password=password, odbc_driver=odbc_driver)


class SqlcmdSQLServerDB(SQLServerDB):
    def __init__(self, host: str = None, instance: str = None, port: int = None, database: str = None,
                 user: str = None, password: str = None, odbc_driver: str = None,
                 protocol: str = None, quoted_identifier: bool = True,
                 trust_server_certificate: bool = False):
        """
        Connection information for a SQL Server database using the MSSQL Tools e.g. sqlcmd

        Args:
            quoted_identifier: If set to true, the SET option QUOTED_IDENTIFIER is set to ON, otherwise OFF.
            protocol: can be tcp (TCP/IP connection), np (named pipe) or lcp (using shared memory).
                      See as well: https://docs.microsoft.com/en-us/sql/ssms/scripting/sqlcmd-connect-to-the-database-engine?view=sql-server-ver15
            trust_server_certificate: Trust the server certificate without validation
        """
        super().__init__(host=host, port=port, database=database, user=user, password=password, odbc_driver=odbc_driver)
        if protocol:
            if protocol not in ['tcp','np','lpc']:
                raise ValueError(f'Not supported protocol: {protocol}')
            if protocol == 'tcp' and instance:
                raise ValueError('You can not use protocol tcp with an instance name')
            if protocol in ['np','lcp'] and port:
                raise ValueError('You can not use protocol np/lcp with a port number')
        if instance is not None and port is not None:
            raise ValueError('You can only use instance or port, not both together')
        self.protocol = protocol
        self.quoted_identifier = quoted_identifier
        self.instance = instance
        self.trust_server_certificate = trust_server_certificate

    @property
    def sqlalchemy_url(self):
        return super().sqlalchemy_url \
                + ('&TrustServerCertificate=yes' if self.trust_server_certificate else '')


class OracleDB(DB):
    def __init__(self, host: str = None, port: int = 0, endpoint: str = None, user: str = None, password: str = None):
        self.host = host
        self.port = port
        self.endpoint = endpoint
        self.user = user
        self.password = password


class SQLiteDB(DB):
    def __init__(self, file_name: pathlib.Path) -> None:
        self.file_name = file_name

    @property
    def sqlalchemy_url(self):
        return f'sqlite:///{self.file_name}'


class SnowflakeDB(DB):
    """A database connection to a Snowflake database"""
    def __init__(self, connection: str = None, account: str = None, user: str = None, password: str = None, database: str = None,
                 private_key_file: str = None, private_key_passphrase: str = None) -> None:
        """
        Connection information for a Snowflake database

        Args:
            connection: The connection name definend in the snowsql configuration ~/.snowsql/config
            account: The account identifier.
                     See here: https://docs.snowflake.com/en/user-guide/admin-account-identifier.html
            user: The user name
            password: The password of the user
            database: The database name
            private_key_file: Path to private key file in PEM format used for key pair authentication.
                              The private key file must be encrypted.
            private_key_passphrase: The passphrase for the private key file.
        """
        self.connection = connection
        self.account = account
        self.user = user
        self.password = password
        self.database = database
        self.private_key_file = private_key_file
        self.private_key_passphrase = private_key_passphrase

    @property
    def sqlalchemy_url(self):
        assert all(v is not None for v in [self.account, self.user, self.password]), \
            "sqlalchemy_url for SnowflakeDB requires a user, password and account"
        return (f'snowflake://{self.user}:{self.password}@{self.account}'
                + (f'/{self.database}' if self.database else ''))


class DatabricksDB(DB):
    """A database connection to a Databricks"""
    def __init__(self, host: str = None, http_path: str = None, access_token: str = None) -> None:
        """
        Connection information for a Databricks

        Args:
            host: The hostname
            http_path: The http path
            access_token: The Access Token
        """
        self.host = host
        self.http_path = http_path
        self.access_token = access_token

    @property
    def sqlalchemy_url(self):
        return f"databricks+connector://token:{self.access_token}@{self.host}:443/"



@functools.singledispatch
def connect(db: object, **kargs) -> object:
    """
    Creating a connection to the database object DB-API 2.0 (PIP-249) compatible.

    See also: https://peps.python.org/pep-0249/#connection-objects

    Args:
        db: The database for which you want to get the database object (either an alias or a `dbs.DB` object)
        **kargs: Optional arguments.
    """
    raise NotImplementedError(f'Please implement connect for type "{db.__class__.__name__}"')


@connect.register(str)
def __(alias: str, **kargs) -> object:
    return connect(db(alias), **kargs)


@connect.register(PostgreSQLDB)
def __(db, **kargs) -> 'psycopg2.extensions.cursor':
    import psycopg2
    return psycopg2.connect(dbname=db.database, user=db.user, password=db.password,
                            host=db.host, port=db.port)


@connect.register(BigQueryDB)
def __(db, **kargs) -> object:
    from google.oauth2.service_account import Credentials
    from google.cloud.bigquery.client import Client
    from google.cloud.bigquery.dbapi.connection import Connection
    credentials = Credentials.from_service_account_file(db.service_account_json_file_name)
    client = Client(project=credentials.project_id, credentials=credentials, location=db.location)
    return Connection(client)


@connect.register(MysqlDB)
def __(db, **kargs) -> 'MySQLdb.cursors.Cursor':
    import MySQLdb.cursors # requires https://github.com/PyMySQL/mysqlclient-python
    return MySQLdb.connect(
        host=db.host, user=db.user, passwd=db.password, db=db.database, port=db.port,
        cursorclass=MySQLdb.cursors.Cursor)


@connect.register(SQLServerDB)
def __(db, **kargs) -> 'pyodbc.Cursor':
    import pyodbc # requires https://github.com/mkleehammer/pyodbc/wiki/Install
    server = db.host
    if db.port: # connecting via TCP/IP port
        server = f"{server},{db.port}"
    return pyodbc.connect(f"DRIVER={{{db.odbc_driver}}};SERVER={server};DATABASE={db.database};UID={db.user};PWD={db.password}" \
                          + (';Encrypt=YES;TrustServerCertificate=YES' if db.trust_server_certificate else ''))


@connect.register(SQLiteDB)
def __(db, **kargs) -> 'sqlite3.Connection':
    import sqlite3
    return sqlite3.connect(database=db.file_name)


@connect.register(DatabricksDB)
def __(db, **kargs) -> object:
    from databricks_dbapi import odbc
    return odbc.connect(
        host=db.host,
        http_path=db.http_path,
        token=db.access_token,
        driver_path=db.odbc_driver_path)



@contextlib.contextmanager
def cursor_context(db: Union[str, DB]) -> object:
    """
    A single iteration with a cursor context. When the iteration is
    closed, a commit is executed on the cursor.

    Example usage:
        with db.cursor_context() as c:
            c.execute('UPDATE table SET table.c1 = 1 WHERE table.id = 5')
    """
    connection = connect(db)
    try:
        cursor = connection.cursor()
        yield cursor
        connection.commit()
    except Exception:
        connection.rollback()
        raise
    finally:
        cursor.close()
        connection.close()



"""Easy access to postgres databases via psycopg2"""

import typing
from warnings import warn

import mara_db.dbs


def postgres_cursor_context(db: typing.Union[str, mara_db.dbs.PostgreSQLDB]) -> 'psycopg2.extensions.cursor':
    """Creates a context with a psycopg2 cursor for a database alias"""
    warn('Function databricks_cursor_context(db) is deprecated. Please use db.cursor_context() instead.')

    if isinstance(db, str):
        db = mara_db.dbs.db(db)

    assert (isinstance(db, mara_db.dbs.PostgreSQLDB))

    return mara_db.dbs.cursor_context(db)



"""Auto-migration of sql alchemy models with alembic. Use with care"""

import copy
import io
import sys
import typing

import sqlalchemy.engine
import sqlalchemy.sql.schema
from sqlalchemy import *  # unfortunately needed to get the eval part further down working
# noinspection PyUnresolvedReferences
from sqlalchemy.dialects import *  # unfortunately needed to get the eval part further down working

import mara_db.dbs
from .sqlalchemy_engine import engine


def auto_migrate(engine: sqlalchemy.engine.Engine, models: typing.List[sqlalchemy.sql.schema.MetaData]):
    """
    Compares a database with a list of defined orm models and applies the diff. Prints executed SQL statements to stdout.

    Based on `alembic automigrations`_, but doesn't require intermediate migration files.

    Use with care, does not work in many cases.

    Args:
        engine: the database to use
        models: A list of orm models

    Returns:
        True in case of no failures

    .. _alembic automigrations:
        http://alembic.zzzcomputing.com/en/latest/autogenerate.html
    """
    import alembic.runtime.migration
    import alembic.autogenerate
    import sqlalchemy_utils

    try:
        # create database if it does not exist
        if not sqlalchemy_utils.database_exists(engine.url):
            sqlalchemy_utils.create_database(engine.url)
            print(f'Created database "{engine.url!r}"\n')
    except Exception as e:
        print(f'Could not access or create database "{engine.url!r}":\n{e}', file=sys.stderr)
        return False

    # merge all models into a single metadata object
    combined_meta_data = MetaData()
    for model in models:
        model.metadata.tables[model.__tablename__].tometadata(combined_meta_data)

    # create diff between models and current db and translate to ddl
    ddl = []
    with engine.connect() as connection:
        output = io.StringIO()

        diff_context = alembic.runtime.migration.MigrationContext(connection.dialect, connection, opts={})

        autogen_context = alembic.autogenerate.api.AutogenContext(diff_context,
                                                                  opts={'sqlalchemy_module_prefix': 'sqlalchemy.',
                                                                        'alembic_module_prefix': 'executor.'})

        execution_context = alembic.runtime.migration.MigrationContext(connection.dialect, connection,
                                                                       opts={'output_buffer': output, 'as_sql': True})

        # needed for the eval below
        executor = alembic.operations.Operations(execution_context)

        # Step 1: create a diff between the meta data and the data base
        # operations is a list of MigrateOperation instances, e.g. a DropTableOp
        operations = alembic.autogenerate.produce_migrations(diff_context, combined_meta_data).upgrade_ops.ops

        for operation in operations:
            # Step 2: autogenerate a python statement from the operation, e.g. "executor.drop_table('bar')"
            renderer = alembic.autogenerate.renderers.dispatch(operation)
            statements = renderer(autogen_context, operation)
            if not isinstance(statements, list):
                statements = [statements]

            for statement in statements:
                # Step 3: "execute" python statement and get sql from buffer, e.g. "DROP TABLE bar;"
                try:
                    eval(statement)
                except Exception as e:
                    print('statement: ' + statement)
                    raise (e)
                ddl.append(output.getvalue())
                output.truncate(0)
                output.seek(0)

    with engine.begin() as connection:
        for statement in ddl:
            sys.stdout.write('\033[1;32m' + statement + '\033[0;0m')
            connection.execute(statement)

    return True


def auto_discover_models_and_migrate() -> bool:
    """
    Auto-migrates all sqlalchemy models that been marked for auto-migration database with the alias 'mara'.

    Models are marked for auto-migration by being put into a module-level `MARA_AUTOMIGRATE_SQLALCHEMY_MODELS`
    variable. E.g.

        MARA_AUTOMIGRATE_SQLALCHEMY_MODELS = [MyModel]

    For this, all modules that contain sqlalchemy models need to be loaded first

    Returns:
        True when no failure happened
    """
    models = []
    for name, module in copy.copy(sys.modules).items():
        if 'MARA_AUTOMIGRATE_SQLALCHEMY_MODELS' in dir(module):
            module_models = getattr(module, 'MARA_AUTOMIGRATE_SQLALCHEMY_MODELS')
            if isinstance(module_models, typing.Callable):
                module_models = module_models()
            if isinstance(models, typing.Dict):
                module_models = module_models.values()
            assert (isinstance(module_models, typing.Iterable))
            models += module_models
    return auto_migrate(engine('mara'), models)


if __name__ == "__main__":
    # Example
    import sqlalchemy.ext.declarative
    import tempfile
    import pathlib

    with tempfile.TemporaryDirectory() as dir:
        db = mara_db.dbs.SQLiteDB(file_name=pathlib.Path(dir) / 'test.sqlite')


        # define a model / table
        class MyTable(sqlalchemy.ext.declarative.declarative_base()):
            __tablename__ = 'my_table'
            my_table_id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
            column_1 = sqlalchemy.Column(sqlalchemy.TEXT, nullable=False, index=True)


        auto_migrate(engine=engine(db), models=[MyTable])


        # ->
        # Created database "sqlite:////var/folders/gg/8117h7rj08zd9rpt55l315_1xx044y/T/tmpl_sdop4j/test.sqlite"
        #
        # CREATE TABLE my_table (
        #     my_table_id SERIAL NOT NULL,
        #     column_1 TEXT NOT NULL,
        #     PRIMARY KEY (my_table_id)
        # );
        #
        # CREATE INDEX ix_my_table_column_1 ON my_table (column_1);

        # remove index and add another column
        class MyTable(sqlalchemy.ext.declarative.declarative_base()):
            __tablename__ = 'my_table'
            my_table_id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
            column_1 = sqlalchemy.Column(sqlalchemy.TEXT, nullable=False)
            column_2 = sqlalchemy.Column(sqlalchemy.Integer)


        auto_migrate(engine=engine(db), models=[MyTable])
        # ->
        # ALTER TABLE my_table ADD COLUMN column_2 INTEGER;
        #
        # DROP INDEX ix_my_table_text_column_1;



"""Auto-migrate command line interface"""

import sys

import click


@click.command()
def migrate():
    """Compares the current database with all defined models and applies the diff"""
    import mara_db.auto_migration

    if not mara_db.auto_migration.auto_discover_models_and_migrate():
        sys.exit(-1)
